import click

from pymetrica.utils import run_profiler

from .codebase_parser import parse_codebase
from .diagram_generator import create_diagram


@click.command()
@click.option(
    "--diagram",
    is_flag=True,
    help="Whether to generate a diagram of the codebase structure.",
)
@click.argument("dir_path")
@run_profiler
def base_stats(dir_path: str, diagram: bool = False) -> None:
    results = parse_codebase(dir_path)
    if diagram:
        create_diagram(results)
    click.echo(
        f"root_folder_path: {results.root_folder_path}\n"
        f"root_folder_name: {results.root_folder_name}\n"
        f"folders_number: {results.folders_number}\n"
        f"files_number: {results.files_number}\n"
        f"lloc_number: {results.lloc_number}\n"
        f"lloc_file_ratio: {results.lloc_file_ratio}\n"
        f"comments_number: {results.comments_number}\n"
        f"comment_line_ratio: {results.comment_lloc_ratio}\n"
        f"classes_number: {results.classes_number}\n"
        f"functions_number: {results.functions_number}",
    )



import ast
from collections.abc import Callable, Generator
from datetime import datetime, timezone
from os import sep
from pathlib import Path
from typing import TypeAlias

from pymetrica.models import Codebase
from pymetrica.utils import log

LayerName: TypeAlias = str
ComponentName: TypeAlias = str
Dependency: TypeAlias = str
Dependencies: TypeAlias = list[Dependency]
Components: TypeAlias = dict[ComponentName, Dependencies]
Layers: TypeAlias = dict[LayerName, Components]


def create_diagram(codebase: Codebase, write: bool = True) -> None:
    layers: Layers = {
        str(dir): Components() for dir in iterdir_generator(codebase.root_folder_path)
    }

    if len(layers) == 0:
        log.warning(f"create_diagram.{codebase.root_folder_path = }")
        layers = {codebase.root_folder_path: Components()}

    for layer_name in layers:
        layers[layer_name].update({
            str(dir): [] for dir in iterdir_generator(layer_name)
        })

    dependencies_visitor = DependenciesVisitor(layers)

    for file in codebase.files:
        if is_root_file(file.filepath, codebase.root_folder_path):
            continue
        for layer in layers:
            if is_component(file.filepath, layer):
                layers[layer][file.filepath] = []
        dependencies_visitor.current_layer = sep.join(
            file.filepath.replace(codebase.root_folder_path + sep, "").split(sep)[:1],
        )
        dependencies_visitor.current_component = file.filepath
        dependencies_visitor.visit(ast.parse(file.code))

    timestamp = datetime.now(tz=timezone.utc).strftime("%Y%m%d_%H%M%S")
    if write:
        with open(f"architecture_diagram_{timestamp}.mmd", "w") as f:  # pylint: disable=unspecified-encoding
            write_diagram(codebase, layers, f.write)
    else:
        write_diagram(codebase, layers, print)


def write_diagram(
    codebase: Codebase,
    layers: Layers,
    write_function: Callable[..., None | int],
) -> None:
    write_function('%%{init: {"themeCSS": ".edgeLabel {font-size: 30px;}"}}%%\n\n')
    write_function("graph TD\n")
    for layer_name, components in layers.items():
        write_function(
            f"  subgraph {layer_name.replace(codebase.root_folder_path + sep, '')}\n",
        )
        for component_name, dependencies in components.items():
            write_function(
                f"    {component_name.replace(codebase.root_folder_path + sep, '')}\n",
            )
        write_function("  end\n")
    for layer_name, components in layers.items():
        for component_name, dependencies in components.items():
            dependencies_count = count_dependencies(dependencies)
            for dep, count in dependencies_count.items():
                write_function(
                    f"  {component_name.replace(codebase.root_folder_path + sep, '')} -- {count} --> {dep.replace(codebase.root_folder_path + sep, '')}\n",  # pylint: disable=line-too-long
                )


def count_dependencies(dependencies: Dependencies) -> dict[str, int]:
    dependencies_count = dict[str, int]()
    for dep in dependencies:
        dependencies_count[dep] = dependencies_count.get(dep, 0) + 1
    return dependencies_count


def iterdir_generator(path: str) -> Generator[str, None, None]:
    for directory in Path(path).iterdir():
        if (
            directory.is_dir()
            and directory.name != "__pycache__"
            and directory.name.startswith(".") is False
        ):
            yield str(directory)


def is_root_file(file_path: str, root_folder_path: str) -> bool:
    relative_path = file_path.replace(root_folder_path, "")
    return relative_path.count(sep) == 1


def is_component(file_path: str, layer_name: str) -> bool:
    relative_path = file_path.replace(layer_name, "")
    return relative_path.count(sep) == 1 and is_python_file_not_dunder(file_path)


def is_python_file_not_dunder(file_path: str) -> bool:
    return not file_path.endswith("__.py") and file_path.endswith(".py")


class DependenciesVisitor(ast.NodeVisitor):
    def __init__(self, layers: Layers) -> None:
        self.layers = layers
        self.layers_names = [layer.rsplit(sep)[-1] for layer in layers]
        self.root_folder = next(iter(layers.keys())).rsplit(sep, 1)[0]
        self.current_layer: str = ""
        self.current_component: str = ""

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:  # pylint: disable=invalid-name
        module_name = node.module if node.module else "relative import"
        split_module_name = module_name.rsplit(".")
        if len(split_module_name) >= 2:
            imported_layer = split_module_name[1]
        else:
            imported_layer = split_module_name[0]
        if (
            split_module_name[0] in self.layers_names
            and split_module_name[0] != self.current_layer
        ):
            imported_layer = split_module_name[0]
        if imported_layer in self.layers_names and imported_layer != self.current_layer:
            full_layer_name_list = [
                layer
                for layer in self.layers
                if layer.rsplit(sep)[-1] == self.current_layer
            ]
            if not full_layer_name_list:
                return
            full_layer_name = full_layer_name_list[0]
            full_imported_layer_name = [
                layer
                for layer in self.layers
                if layer.rsplit(sep)[-1] == imported_layer
            ]
            subdirectory_path = self.current_component.replace(
                self.root_folder,
                "",
            ).split(sep, 1)[-1]
            if subdirectory_path.count(sep) == 1:
                clean_current_component = self.root_folder + sep + subdirectory_path
            else:
                clean_subdirectory_path = sep.join(subdirectory_path.split(sep, 2)[:2])
                clean_current_component = (
                    self.root_folder + sep + clean_subdirectory_path
                )
            try:
                if self.layers[full_layer_name].get(clean_current_component):
                    self.layers[full_layer_name][clean_current_component].append(
                        full_imported_layer_name,  # type: ignore[arg-type]
                    )
                else:
                    self.layers[full_layer_name].update({
                        clean_current_component: [full_imported_layer_name],  # type: ignore[list-item]  # pylint: disable=line-too-long
                    })
            except KeyError as e:
                log.warning(
                    f"DependenciesVisitor.visit_ImportFrom.KeyError.{clean_current_component = }",
                )
                log.warning(f"DependenciesVisitor.visit_ImportFrom.KeyError: {e = }")



import click

from pymetrica.codebase_parser import parse_codebase
from pymetrica.metric_calculators.abstract_lines_of_code.aloc_calculator import (
    AlocCalculator,
)
from pymetrica.metric_calculators.cyclomatic_complexity.cc_calculator import (
    CCCalculator,
)
from pymetrica.metric_calculators.halstead_volume.hv_calculator import (
    HalsteadVolumeCalculator,
)
from pymetrica.metric_calculators.instability.instability_calculator import (
    InstabilityCalculator,
)
from pymetrica.metric_calculators.maintainability_cost.mc_calculator import (
    MaintainabilityCostCalculator,
)
from pymetrica.models.metric import Metric, Results
from pymetrica.report_generators.reports_mapping import REPORTS_MAPPING

aloc_calculator: AlocCalculator = AlocCalculator()
cc_calculator: CCCalculator = CCCalculator()
hv_calculator: HalsteadVolumeCalculator = HalsteadVolumeCalculator()
mc_calculator: MaintainabilityCostCalculator = MaintainabilityCostCalculator()
instability_calculator: InstabilityCalculator = InstabilityCalculator()


@click.command()
@click.option(
    "--long-report",
    is_flag=True,
    help="Whether to generate a long summary report for each metric.",
)
@click.argument("dir_path")
@click.option("-rt", "--report-type", type=str, default="BASIC_TERMINAL")
def run_all(
    dir_path: str,
    report_type: str,
    long_report: bool = False,
) -> None:
    results = parse_codebase(dir_path)
    metrics = list[Metric[Results]]()
    metrics.append(aloc_calculator.calculate_metric(results))
    metrics.append(cc_calculator.calculate_metric(results))
    metrics.append(hv_calculator.calculate_metric(results))
    metrics.append(mc_calculator.calculate_metric(results))
    metrics.append(instability_calculator.calculate_metric(results))

    report_generator = REPORTS_MAPPING[report_type]()
    if long_report:
        click.echo(report_generator.generate_report(metrics))
        return
    click.echo(report_generator.generate_short_report(metrics))



import ast
from os import sep

from pymetrica.models.code import Code

Layer = str
Files = list[Code]


def clean_layer_name(layer_path: str) -> str:
    return " " + layer_path.rsplit(sep)[-2] + "." + layer_path.rsplit(sep)[-1]


def calculate_instability(
    analyzed_layer: Layer,
    files_by_layer: dict[Layer, Files],
) -> float:
    analyzed_layer = clean_layer_name(analyzed_layer)
    all_layers = [clean_layer_name(layer) for layer in files_by_layer]
    efferent_coupling = 0
    afferent_coupling = 0
    for layer in files_by_layer:  # noqa: PLC0206
        instability_visitor = InstabilityVisitor(
            analyzed_layer,
            clean_layer_name(layer),
            all_layers,
        )
        for file in files_by_layer[layer]:
            instability_visitor.visit(ast.parse(file.code))

        efferent_coupling += instability_visitor.efferent_imports
        afferent_coupling += instability_visitor.afferent_imports

    if efferent_coupling == 0:
        return 0.0

    return efferent_coupling / (efferent_coupling + afferent_coupling)


class InstabilityVisitor(ast.NodeVisitor):
    def __init__(
        self,
        analyzed_layer: Layer,
        traversing_layer: Layer,
        all_layers: list[Layer],
    ) -> None:
        self.analyzed_layer = analyzed_layer
        self.all_layers = all_layers
        self.efferent_imports = 0
        self.afferent_imports = 0

        if analyzed_layer == traversing_layer:
            self.look_for_efferent = True
        else:
            self.look_for_efferent = False

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:  # pylint: disable=invalid-name
        module_name = node.module if node.module else "relative import"
        module_name = " " + module_name + " "
        if self.look_for_efferent:
            for layer in self.all_layers:
                if layer != self.analyzed_layer and layer in module_name:
                    self.efferent_imports += 1
                    break
        if not self.look_for_efferent and self.analyzed_layer in module_name:
            self.afferent_imports += 1
        self.generic_visit(node)



from typing import Any

from pydantic import BaseModel

from pymetrica.models import Metric, Results


class LayerHV(BaseModel):
    name: str
    hv_number: float


class HalsteadVolumeResults(Results):
    hv_number: float
    hv_per_layer: list[LayerHV]

    def get_dict(self) -> dict[str, Any]:
        return self.model_dump(exclude={"hv_per_layer"})

    def get_json(self) -> str:
        return self.get_json()

    def get_summary(self) -> str:
        summary = f"\nHalstead Volume: {self.hv_number:.2f}\n"
        for layer in self.hv_per_layer:
            summary += f"  Layer {layer.name}: {layer.hv_number:.2f}\n"
        return summary


class HalsteadVolumeMetric(Metric[HalsteadVolumeResults]): ...



import ast
import math
import os

from pymetrica.models import Codebase, MetricCalculator

from .hv_metric import HalsteadVolumeMetric, HalsteadVolumeResults, LayerHV
from .hv_visitor import HalsteadVolumeVisitor


class HalsteadVolumeCalculator(MetricCalculator[HalsteadVolumeResults]):
    def calculate_metric(  # pylint: disable=too-many-locals
        self: "HalsteadVolumeCalculator",
        codebase: Codebase,
    ) -> HalsteadVolumeMetric:
        layer_results = list[LayerHV]()
        layers = codebase.layers.copy()
        layers.update({"root": codebase.root_files})

        codebase_unique_operators = set[str]()
        codebase_unique_operands = set[str]()
        codebase_operators = 0
        codebase_operands = 0

        for layer_name, layer_files in layers.items():
            total_hv = 0.0

            for code_file in layer_files:
                tree = ast.parse(code_file.code)
                visitor = HalsteadVolumeVisitor()
                visitor.visit(tree)

                codebase_unique_operators.update(set(visitor.operators))
                codebase_unique_operands.update(set(visitor.operands))
                codebase_operators += len(visitor.operators)
                codebase_operands += len(visitor.operands)
                unique_operators = len(set(visitor.operators))
                unique_operands = len(set(visitor.operands))
                operators = len(visitor.operators)
                operands = len(visitor.operands)

                if unique_operators + unique_operands == 0:
                    halstead_volume = 0.0
                else:
                    vocabulary = unique_operators + unique_operands
                    length = operators + operands
                    halstead_volume = length * math.log2(vocabulary)

                total_hv += halstead_volume

            layer_results.append(
                LayerHV(
                    name=layer_name.rsplit(os.sep, 1)[-1],
                    hv_number=total_hv,
                ),
            )

        count_of_unique_operators = len(codebase_unique_operators)
        count_of_unique_operands = len(codebase_unique_operands)

        if count_of_unique_operators + count_of_unique_operands == 0:
            codebase_halstead_volume = 0.0
        else:
            vocabulary = count_of_unique_operators + count_of_unique_operands
            length = codebase_operators + codebase_operands
            codebase_halstead_volume = length * math.log2(vocabulary)

        return HalsteadVolumeMetric(
            name="Halstead Volume",
            description=(""),
            results=HalsteadVolumeResults(
                hv_number=codebase_halstead_volume,
                hv_per_layer=layer_results,
            ),
        )



from collections.abc import Callable
from cProfile import Profile
from functools import wraps
from io import StringIO
from pstats import SortKey, Stats
from typing import ParamSpecArgs, ParamSpecKwargs, TypeVar

from .logging import log

profiler = Profile()

T = TypeVar("T")


def run_profiler(func: Callable[..., T]) -> Callable[..., T]:
    @wraps(func)
    def wrapper(*args: ParamSpecArgs, **kwargs: ParamSpecKwargs) -> T:
        profiler.enable()
        result = func(*args, **kwargs)
        profiler.disable()
        temp_output = StringIO()
        stats = Stats(profiler, stream=temp_output).sort_stats(SortKey.CUMULATIVE)
        stats.print_stats(7)
        log.debug(temp_output.getvalue())
        return result

    return wrapper



from .get_hv import hv
from .hv_calculator import HalsteadVolumeCalculator
from .hv_metric import HalsteadVolumeMetric, HalsteadVolumeResults

__all__ = [
    "HalsteadVolumeCalculator",
    "HalsteadVolumeMetric",
    "HalsteadVolumeResults",
    "hv",
]



from .get_instability import li
from .instability_calculator import InstabilityCalculator
from .instability_metric import InstabilityMetric, InstabilityResults

__all__ = [
    "InstabilityCalculator",
    "InstabilityMetric",
    "InstabilityResults",
    "li",
]



import os

# the following imports need to be more specific to avoid cyclic imports
from pymetrica.metric_calculators.cyclomatic_complexity import (
    CCCalculator,
)
from pymetrica.metric_calculators.halstead_volume import (
    HalsteadVolumeCalculator,
)
from pymetrica.models import Codebase, Metric, MetricCalculator
from pymetrica.utils import log

from .mc_metric import LayerMC, MaintainabilityCostMetric, MaintainabilityCostResults


class MaintainabilityCostCalculator(MetricCalculator[MaintainabilityCostResults]):
    def calculate_metric(  # pylint: disable=too-many-locals
        self: "MaintainabilityCostCalculator",
        codebase: Codebase,
    ) -> Metric[MaintainabilityCostResults]:
        layers = codebase.layers.copy()
        layers.update({"root": codebase.root_files})
        cc_calculator = CCCalculator()
        hv_calculator = HalsteadVolumeCalculator()
        cc_metric = cc_calculator.calculate_metric(codebase)
        hv_metric = hv_calculator.calculate_metric(codebase)

        layers_results = list[LayerMC]()
        for layer_name, layer_files in layers.items():
            layer_name = layer_name.rsplit(os.sep, 1)[-1]
            layer_lloc = sum(file.lloc_number for file in layer_files)
            layer_cc = [  # noqa: RUF015
                result.cc_number
                for result in cc_metric.results.cc_result_per_layer
                if result.name == layer_name
            ][0]
            layer_hv = [  # noqa: RUF015
                result.hv_number
                for result in hv_metric.results.hv_per_layer
                if result.name == layer_name
            ][0]

            hv_density = layer_hv / (layer_lloc or 1)
            cc_density = layer_cc / (layer_lloc or 1)
            average_lloc_mc = (hv_density * ((cc_density) * 100 or 1)) / 20
            log.debug(
                f"Layer: {layer_name}, HV Density: {hv_density}, "
                f"CC Density: {cc_density}, Average LLOC MC: {average_lloc_mc}",
            )
            mc = average_lloc_mc + layer_lloc * 0.001
            layers_results.append(
                LayerMC(
                    name=layer_name,
                    maintainability_cost=mc,
                    raw_line_cost=average_lloc_mc,
                ),
            )

        codebase_hv_density = hv_metric.results.hv_number / (codebase.lloc_number or 1)
        codebase_cc_density = cc_metric.results.cc_number / (codebase.lloc_number or 1)
        codebase_average_lloc_mc = (
            codebase_hv_density * ((codebase_cc_density) * 100 or 1)
        ) / 20
        log.debug(
            f"Codebase HV Density: {codebase_hv_density}, "
            f"CC Density {codebase_cc_density}, Average LLOC MC: {codebase_average_lloc_mc}",
        )
        codebase_mc = codebase_average_lloc_mc + codebase.lloc_number * 0.001

        return MaintainabilityCostMetric(
            name="Maintainability Cost",
            description=(
                "MC is a software metric that measures how expensive it is to "
                "maintain the code, based on various factors such as "
                "Cyclomatic Complexity, Logical Lines Of Code, and Halstead "
                "Volume. Lower scores indicate better maintainability, with "
                "scores above 20 suggesting moderate maintainability and "
                "scores above 50 indicating poor maintainability."
            ),
            results=MaintainabilityCostResults(
                maintainability_cost=codebase_mc,
                raw_line_cost=codebase_average_lloc_mc,
                mc_per_layer=layers_results,
            ),
        )



import click

from pymetrica.codebase_parser import parse_codebase
from pymetrica.report_generators import REPORTS_MAPPING
from pymetrica.utils import run_profiler

from .mc_calculator import MaintainabilityCostCalculator

mc_calculator: MaintainabilityCostCalculator = MaintainabilityCostCalculator()


@click.command()
@click.argument("dir_path")
@click.option("-rt", "--report-type", type=str, default="BASIC_TERMINAL")
@run_profiler
def mc(
    dir_path: str,
    report_type: str,
) -> None:
    codebase = parse_codebase(dir_path)
    mc_metric = mc_calculator.calculate_metric(codebase)
    report_generator = REPORTS_MAPPING[report_type]()
    click.echo(report_generator.generate_report([mc_metric]))



from .abstract_lines_of_code import AlocCalculator, aloc
from .cyclomatic_complexity import CCCalculator, cc
from .halstead_volume import HalsteadVolumeCalculator, hv
from .instability import (
    InstabilityCalculator,
    InstabilityResults,
    li,
)
from .maintainability_cost import (
    MaintainabilityCostCalculator,
    MaintainabilityCostResults,
    mc,
)

__all__ = [
    "AlocCalculator",
    "CCCalculator",
    "HalsteadVolumeCalculator",
    "InstabilityCalculator",
    "InstabilityResults",
    "MaintainabilityCostCalculator",
    "MaintainabilityCostResults",
    "aloc",
    "cc",
    "hv",
    "li",
    "mc",
]



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from .codebase import Codebase
from .metric import Metric, Results

T = TypeVar("T", bound=Results)


class MetricCalculator(ABC, Generic[T]):
    @abstractmethod
    def calculate_metric(self, codebase: Codebase) -> Metric[T]:
        raise NotImplementedError(
            "Subclasses of MetricCalculator must implement this method.",
        )



# pylint: disable=invalid-name

import ast

# This has been based on Radon's implementation and updated to support
# modern python features.


class HalsteadVolumeVisitor(ast.NodeVisitor):
    def __init__(self) -> None:
        self.operators: list[str] = []
        self.operands: list[str] = []

    def visit_BinOp(self, node: ast.BinOp) -> None:
        # Binary operators (+, -, *, etc.)
        self.operators.append(type(node.op).__name__)
        self.generic_visit(node)

    def visit_UnaryOp(self, node: ast.UnaryOp) -> None:
        # Unary operators (+x, -x, not x)
        self.operators.append(type(node.op).__name__)
        self.generic_visit(node)

    def visit_BoolOp(self, node: ast.BoolOp) -> None:
        # Boolean operators (and, or)
        self.operators.append(type(node.op).__name__)
        self.generic_visit(node)

    def visit_Compare(self, node: ast.Compare) -> None:
        # Comparison operators (<, >, ==, etc.)
        for op in node.ops:
            self.operators.append(type(op).__name__)
        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        # Assignment
        self.operators.append("Assign")
        self.generic_visit(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        # Augmented assignment (+=, -=, etc.)
        self.operators.append(f"AugAssign_{type(node.op).__name__}")
        self.generic_visit(node)

    def visit_If(self, node: ast.If) -> None:
        self.operators.append("if")
        self.generic_visit(node)

    def visit_For(self, node: ast.For) -> None:
        self.operators.append("for")
        self.generic_visit(node)

    def visit_While(self, node: ast.While) -> None:
        self.operators.append("while")
        self.generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.operators.append("def")
        self.operands.append(node.name)  # function name is an operand
        self.generic_visit(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self.operators.append("class")
        self.operands.append(node.name)  # class name is an operand
        self.generic_visit(node)

    def visit_Name(self, node: ast.Name) -> None:
        # Variable name
        self.operands.append(node.id)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        # Object attribute
        self.operands.append(node.attr)
        self.generic_visit(node)

    def visit_Constant(self, node: ast.Constant) -> None:
        # Literals
        self.operands.append(repr(node.value))



import os

from pymetrica.utils import is_logical_line_of_code


def count_logical_lines_of_code(root: str, file_name: str) -> int:
    path = os.path.join(root, file_name)
    with open(path, encoding="utf-8") as f:
        lines = f.readlines()
    return sum(1 for line in lines if is_logical_line_of_code(line))



from collections.abc import Sequence
from typing import TypeVar

from pymetrica.models import Metric, ReportGenerator, Results

T = TypeVar("T", bound=Results)


class BasicTerminalReport(ReportGenerator):
    def generate_report(self, metrics: Sequence[Metric[T]]) -> str:
        report = ""
        for metric in metrics:
            report += f"Metric: {metric.name}\n"
            report += f"Description: {metric.description}\n"
            report += f"Summary: {metric.results.get_summary()}\n"
            report += "-" * 100
            report += "\n"
        return report

    def generate_short_report(self, metrics: Sequence[Metric[T]]) -> str:
        report = "-" * 100
        report += "\nShort Report\n"
        report += "-" * 100
        for metric in metrics:
            report += f"\nMetric: {metric.name}\n"
            results = metric.results.get_dict()
            for key, value in results.items():
                report += f"{key}: {value}\n"
            report += "-" * 100
        return report



from pydantic import BaseModel

from .code import Code


class Codebase(BaseModel):
    root_folder_path: str
    root_folder_name: str
    folders_number: int
    files_number: int
    lloc_number: int
    lloc_file_ratio: str
    comments_number: int
    comment_lloc_ratio: str
    classes_number: int
    functions_number: int
    layers: dict[str, list[Code]]
    root_files: list[Code]

    @property
    def files(self) -> list[Code]:
        return self.root_files + [
            file for layer in self.layers.values() for file in layer
        ]



from abc import ABC, abstractmethod
from typing import TypeVar

from .metric import Metric, Results

T = TypeVar("T", bound=Results)


class ReportGenerator(ABC):
    @abstractmethod
    def generate_report(self, metrics: list[Metric[T]]) -> str:
        raise NotImplementedError(
            "Subclasses of ReportGenerator must implement this method.",
        )



import os
import sys
from enum import Enum

from loguru import logger
from pydantic import TypeAdapter


class LogLevel(Enum):
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    DEBUG = "DEBUG"


LogLevelAdapter = TypeAdapter(LogLevel)

LOG_LEVEL = os.getenv("PYMETRICA_LOG_LEVEL")
try:
    log_level = LogLevelAdapter.validate_python(LOG_LEVEL)
    logger.info(f"Environment log level set to: {log_level.value}")
except ValueError:
    log_level = LogLevel.INFO
    logger.warning(
        f"Invalid environment log level '{LOG_LEVEL}' provided. Defaulting to '{log_level.value}'.",
    )


class CustomLogger:
    def __init__(self) -> None:
        logger.remove()
        self.level = log_level.value
        logger.add(sys.stderr, level=self.level)
        self.logger = logger

    def info(self, message: str) -> None:
        self.logger.info(message)

    def warning(self, message: str) -> None:
        self.logger.warning(message)

    def error(self, message: str) -> None:
        self.logger.error(message)

    def debug(self, message: str) -> None:
        self.logger.debug(message)


log: CustomLogger = CustomLogger()

log.info(f"Pymetrica logging initialized with level: {log.level}")



import ast

from pydantic import BaseModel

from pymetrica.models import Code
from pymetrica.utils import is_logical_line_of_code, log


class PreliminaryResults(BaseModel):
    aloc: int
    classes: dict[str, int]


def gather_loc_and_classes(files: list[Code]) -> PreliminaryResults:  # pylint: disable=too-many-locals, too-many-branches
    """
    First pass:
      - Count total logical lines ('lloc').
      - Identify all class definitions and compute their LOC spans.
      - Compute abstract LOC ('aloc').
      - Record non-abstract class LOC in 'classes'.
    """
    total_lloc = 0
    total_aloc = 0
    classes: dict[str, int] = {}

    for file in files:  # pylint: disable=too-many-nested-blocks
        tree = ast.parse(file.code)

        # Count logical lines in file
        file_lloc = sum(1 for line in file.code_lines if is_logical_line_of_code(line))
        total_lloc += file_lloc

        # Collect abstract lines
        file_abs: set[int] = set()
        for node in ast.walk(tree):
            # Imports
            if isinstance(node, (ast.Import, ast.ImportFrom)):
                file_abs.add(node.lineno)
            # __all__ assignments
            if isinstance(node, ast.Assign):
                for tgt in node.targets:
                    if isinstance(tgt, ast.Name) and tgt.id == "__all__":
                        file_abs.add(node.lineno)
            # Function defs and decorators
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                file_abs.add(node.lineno)
                for dec in node.decorator_list:
                    file_abs.add(dec.lineno)
            # Class defs and decorators
            if isinstance(node, ast.ClassDef):
                file_abs.add(node.lineno)
                for dec in node.decorator_list:
                    file_abs.add(dec.lineno)
            # # Any call TODO this was removed because many calls may be done in one single line
            # if isinstance(node, ast.Call):
            #     file_abs.add(node.lineno)

        total_aloc += len(file_abs)

        # Measure classes
        for node in tree.body:
            if isinstance(node, ast.ClassDef):
                start, end = (
                    node.lineno,
                    getattr(
                        node,
                        "end_lineno",
                        node.lineno,
                    ),
                )
                # Check inheritance
                bases: list[str] = []
                for b in node.bases:
                    if isinstance(b, ast.Name):
                        bases.append(b.id)
                    else:
                        try:
                            bases.append(ast.unparse(b))
                        except:  # noqa: E722  # pylint: disable=bare-except
                            log.warning(
                                f"gather_loc_and_classes.unparse_failed.{ast.dump(b) = }",
                            )
                # Treat ABC, ABCMeta, Protocol, and Enum as abstract
                is_abstract_base = any(
                    base in ("ABC", "ABCMeta", "Protocol", "Enum") for base in bases
                )
                # Count LOC
                loc = sum(
                    1
                    for i in range(start - 1, end)
                    if i < len(file.code_lines)
                    and is_logical_line_of_code(file.code_lines[i])
                )
                if not is_abstract_base:
                    # -1 to exclude class definition line
                    classes[node.name] = loc - 1

    return PreliminaryResults(aloc=total_aloc, classes=classes)



from .get_mc import mc
from .mc_calculator import MaintainabilityCostCalculator
from .mc_metric import MaintainabilityCostMetric, MaintainabilityCostResults

__all__ = [
    "MaintainabilityCostCalculator",
    "MaintainabilityCostMetric",
    "MaintainabilityCostResults",
    "mc",
]



from typing import Any

from pydantic import BaseModel

from pymetrica.models import Metric, Results


class LayerAloc(BaseModel):
    name: str
    aloc_number: int
    aloc_percentage: float


class AlocResults(Results):
    aloc_number: int
    aloc_percentage: float
    aloc_result_per_layer: list[LayerAloc]

    def get_dict(self) -> dict[str, Any]:
        return self.model_dump(exclude={"aloc_result_per_layer"})

    def get_json(self) -> str:
        return self.get_json()

    def get_summary(self) -> str:
        summary = (
            "\nTotal ALOC: "
            f"{self.aloc_number} ({self.aloc_percentage:0.2f}% of total LLOC)\n"
        )
        for layer in self.aloc_result_per_layer:
            summary += (
                f"  Layer {layer.name} ALOC: "
                f"{layer.aloc_number} ({layer.aloc_percentage:0.2f}%)\n"
            )
        return summary


class AlocMetric(Metric[AlocResults]): ...



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, ConfigDict


class Results(BaseModel, ABC):
    @abstractmethod
    def get_json(self) -> str:
        raise NotImplementedError(
            "Subclasses of Results must implement this method.",
        )

    @abstractmethod
    def get_dict(self) -> dict[str, Any]:
        raise NotImplementedError(
            "Subclasses of Results must implement this method.",
        )

    @abstractmethod
    def get_summary(self) -> str:
        raise NotImplementedError(
            "Subclasses of Results must implement this method.",
        )


T_co = TypeVar("T_co", bound=Results, covariant=True)


class Metric(BaseModel, Generic[T_co]):
    name: str
    description: str
    results: T_co

    model_config = ConfigDict(arbitrary_types_allowed=True)



PARENS = ["(", ")", "{", "}", "[", "]", ",", ":"]


def is_logical_line_of_code(line: str) -> bool:
    """Check whether a line is blank or a comment."""
    stripped = line.strip()
    if not stripped:
        return False
    if stripped.startswith("#"):
        return False
    for char in stripped:
        if char not in PARENS:
            break
    else:
        return False
    return True



import ast
import os
from pathlib import Path

from pymetrica.models import Code, Codebase
from pymetrica.utils import is_comment_line, is_logical_line_of_code, log

# TODO ignore folders inside .gitignore


def parse_codebase(dir_path: str) -> Codebase:  # pylint: disable=too-many-locals
    base = Path(dir_path).absolute()
    total_lloc = 0
    total_comments = 0
    total_classes_definitions = 0
    total_functions_definitions = 0
    layers = get_layers_from_path(base)
    root_files = list[Code]()

    for path in base.rglob("*.py"):
        source = path.read_text(encoding="utf-8")
        lines = source.splitlines(keepends=True)
        try:
            tree = ast.parse(source)
        except SyntaxError as e:
            log.warning(f"parse_codebase.SyntaxError.{path = }")
            log.warning(f"parse_codebase.SyntaxError: {e = }")
            continue
        total_classes_definitions += sum(
            isinstance(n, ast.ClassDef) for n in ast.walk(tree)
        )
        total_functions_definitions += sum(
            isinstance(n, (ast.FunctionDef, ast.AsyncFunctionDef))
            for n in ast.walk(tree)
        )

        total_lloc += (
            file_lloc := sum(1 for line in lines if is_logical_line_of_code(line))
        )
        total_comments += (
            file_comments := sum(1 for line in lines if is_comment_line(line))
        )
        list_reference = None
        for layer in layers:
            if str(path).startswith(layer):
                list_reference = layers[layer]
                break
        if list_reference is None:
            list_reference = root_files
        list_reference.append(
            Code(
                filepath=str(path),
                filename=path.name,
                lloc_number=file_lloc,
                comments_number=file_comments,
                code_lines=lines,
                code=source,
            ),
        )

    total_files = sum(len(files) for files in layers.values()) + len(root_files)

    return Codebase(
        root_folder_path=str(Path(dir_path).absolute()),
        root_folder_name=os.path.basename(dir_path),
        folders_number=sum(1 for p in base.rglob("*") if p.is_dir()),
        files_number=total_files,
        lloc_number=total_lloc,
        lloc_file_ratio=(
            f"{total_lloc / (total_files or 1):.1f}:{1 if total_files else 0}"
        ),
        comments_number=total_comments,
        comment_lloc_ratio=(
            f"{1 if total_comments else 0}:{total_lloc / (total_comments or 1):.1f}"
        ),
        classes_number=total_classes_definitions,
        functions_number=total_functions_definitions,
        layers=layers,
        root_files=root_files,
    )


def get_layers_from_path(base: Path) -> dict[str, list[Code]]:
    return {
        str(dir): list[Code]()
        for dir in base.iterdir()
        if dir.is_dir() and dir.name != "__pycache__"
    }



import json

from pymetrica.models import Metric, Results


class InstabilityResults(Results):
    instability: dict[str, float]

    def get_dict(self) -> dict[str, float]:
        return self.instability

    def get_json(self) -> str:
        return json.dumps(self.get_dict())

    def get_summary(self) -> str:
        summary = "\nInstability for:\n"
        for layer, instability in self.instability.items():
            summary += f"  {layer}: {instability}\n"
        return summary


class InstabilityMetric(Metric[InstabilityResults]): ...



from typing import Any

from pydantic import BaseModel

from pymetrica.models import Metric, Results


class LayerCC(BaseModel):
    name: str
    cc_number: int
    lloc_per_cc: float


class CCResults(Results):
    cc_number: int
    lloc_per_cc: float
    cc_result_per_layer: list[LayerCC]

    def get_dict(self) -> dict[str, Any]:
        return self.model_dump(exclude={"cc_result_per_layer"})

    def get_json(self) -> str:
        return self.get_json()

    def get_summary(self) -> str:
        summary = (
            f"\nTotal CC: {self.cc_number} ({self.lloc_per_cc:0.2f} LLOC per CC)\n"
        )
        for layer in self.cc_result_per_layer:
            summary += (
                f"  Layer {layer.name} CC: "
                f"{layer.cc_number} ({layer.lloc_per_cc:0.2f} LLOC per CC)\n"
            )
        return summary


class CCMetric(Metric[CCResults]): ...



from .code import Code
from .codebase import Codebase
from .metric import Metric, Results
from .metric_calculator import MetricCalculator
from .report_generator import ReportGenerator

__all__ = [
    "Code",
    "Codebase",
    "Metric",
    "MetricCalculator",
    "ReportGenerator",
    "Results",
]



from os import sep

from pymetrica.models import Code, Codebase, Metric, MetricCalculator

from .instability_logic import calculate_instability
from .instability_metric import InstabilityMetric, InstabilityResults

Layer = str
Files = list[Code]


def get_layer_name(layer_path: str) -> str:
    return layer_path.rsplit(sep)[-1]


class InstabilityCalculator(MetricCalculator[InstabilityResults]):
    def calculate_metric(
        self: "InstabilityCalculator",
        codebase: Codebase,
    ) -> Metric[InstabilityResults]:
        codebase.layers.update({codebase.root_folder_path: codebase.root_files})
        files_by_layer: dict[Layer, Files] = codebase.layers.copy()

        layers_instability = dict[Layer, float]()

        for layer in files_by_layer:
            if layer == codebase.root_folder_path:
                layers_instability["root"] = calculate_instability(
                    layer,
                    files_by_layer,
                )
                continue
            layers_instability[get_layer_name(layer)] = calculate_instability(
                layer,
                files_by_layer,
            )

        return InstabilityMetric(
            name="Instability",
            description=(
                "The Instability metric measures the coupling and stability of each layer. "
                "It is calculated based on the ratio of outgoing dependencies to the total "
                "number of dependencies (incoming + outgoing). A higher value indicates "
                "greater instability."
            ),
            results=InstabilityResults(
                instability=layers_instability,
            ),
        )



import os

from pymetrica.models import Codebase, MetricCalculator

from .aloc_metric import AlocMetric, AlocResults, LayerAloc
from .first_pass import gather_loc_and_classes
from .second_pass import called_classes


class AlocCalculator(MetricCalculator[AlocResults]):
    def calculate_metric(self: "AlocCalculator", codebase: Codebase) -> AlocMetric:
        layer_results = list[LayerAloc]()
        layers = codebase.layers.copy()
        layers.update({"root": codebase.root_files})
        all_called_classes = set[str]()
        classes_per_layer = dict[str, dict[str, int]]()

        for name, files in layers.items():
            preliminary_results = gather_loc_and_classes(files)
            classes_per_layer[name] = {}
            classes_per_layer[name].update(preliminary_results.classes)
            layer_results.append(
                LayerAloc(
                    name=name.rsplit(os.sep, 1)[-1],
                    aloc_number=preliminary_results.aloc,
                    aloc_percentage=preliminary_results.aloc
                    / codebase.lloc_number
                    * 100,
                ),
            )
        all_classes = dict[str, int]()
        for classes in classes_per_layer.values():
            all_classes.update(classes)
        all_called_classes = called_classes(codebase.files, all_classes)
        for layer in classes_per_layer:
            for layer_result in layer_results:
                if layer_result.name == layer:
                    layer_result.aloc_number += sum(
                        loc
                        for cls, loc in classes_per_layer[layer].items()
                        if cls not in all_called_classes
                    )
                    layer_result.aloc_percentage = (
                        layer_result.aloc_number / codebase.lloc_number * 100
                    )

        total_aloc = sum(layer.aloc_number for layer in layer_results)
        return AlocMetric(
            name="Abstract Lines Of Code",
            description=(
                "Abstract Lines Of Code (ALOC) is a software metric that "
                "measures the number of lines of code in a program that are "
                "not concrete operations but rather indirections or abstract "
                "constructs and definitions."
            ),
            results=AlocResults(
                aloc_number=total_aloc,
                aloc_percentage=total_aloc / codebase.lloc_number * 100,
                aloc_result_per_layer=sorted(
                    layer_results,
                    key=lambda x: x.aloc_number,
                    reverse=True,
                ),
            ),
        )



import ast
import os

from pymetrica.models import Codebase, MetricCalculator

from .cc_metric import CCMetric, CCResults, LayerCC
from .cc_visitor import CCVisitor


class CCCalculator(MetricCalculator[CCResults]):
    def calculate_metric(self: "CCCalculator", codebase: Codebase) -> CCMetric:
        layer_results = list[LayerCC]()
        layers = codebase.layers.copy()
        layers.update({"root": codebase.root_files})

        codebase_complexity = 1

        for layer_name, layer_files in layers.items():
            layer_complexity = 0
            layer_lloc = 0

            for code_file in layer_files:
                tree = ast.parse(code_file.code)
                visitor = CCVisitor()
                visitor.visit(tree)
                layer_complexity += visitor.complexity
                layer_lloc += code_file.lloc_number

            layer_results.append(
                LayerCC(
                    name=layer_name.rsplit(os.sep, 1)[-1],
                    cc_number=layer_complexity,
                    lloc_per_cc=layer_lloc / layer_complexity
                    if layer_complexity > 0
                    else 0,
                ),
            )
            codebase_complexity += layer_complexity

        return CCMetric(
            name="Cyclomatic Complexity",
            description=(
                "Cyclomatic Complexity (CC) is a software metric used to "
                "measure the complexity of a program. It is calculated based "
                "on the control flow graph of the program, where nodes "
                "represent code blocks and edges represent control flow paths. "
            ),
            results=CCResults(
                cc_number=codebase_complexity,
                lloc_per_cc=codebase.lloc_number / codebase_complexity
                if codebase_complexity > 0
                else 0,
                cc_result_per_layer=layer_results,
            ),
        )



import ast

from pymetrica.models import Code


def called_classes(files: list[Code], classes: dict[str, int]) -> set[str]:
    """
    Second pass:
      - Detect instantiation by:
        * Direct calls: ClassName()
        * Class method calls: ClassName.method(...)
      - Any class appearing as ast.Name in call.func or ast.Attribute.value qualifies.
    Returns sum of LOC for classes never instantiated.
    """
    called: set[str] = set()
    for file in files:
        tree = ast.parse(file.code)
        for node in ast.walk(tree):
            if isinstance(node, ast.Call):
                func = node.func
                # Direct instantiation
                if isinstance(func, ast.Name) and func.id in classes:
                    called.add(func.id)
                # Method call on class (ClassName.method(...))
                elif isinstance(func, ast.Attribute):
                    val = func.value
                    if isinstance(val, ast.Name) and val.id in classes:
                        called.add(val.id)
                for arg in node.args:
                    if isinstance(arg, ast.Name) and arg.id in classes:
                        called.add(arg.id)
    return called



import click

from pymetrica.metric_calculators import (
    aloc,
    cc,
    hv,
    li,
    mc,
)

from .codebase_parser import base_stats
from .run_all import run_all
from .status import status


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
def main() -> None: ...


main.add_command(status)
main.add_command(run_all)
main.add_command(base_stats)
main.add_command(aloc)
main.add_command(cc)
main.add_command(hv)
main.add_command(mc)
main.add_command(li)

if __name__ == "__main__":
    main()



from .is_comment import is_comment_line
from .is_lloc import is_logical_line_of_code
from .logging import log
from .profiler import run_profiler

__all__ = ["is_comment_line", "is_logical_line_of_code", "log", "run_profiler"]



import json

from pydantic import BaseModel

from pymetrica.models import Metric, Results


class LayerMC(BaseModel):
    name: str
    maintainability_cost: float
    raw_line_cost: float


class MaintainabilityCostResults(Results):
    maintainability_cost: float
    raw_line_cost: float
    mc_per_layer: list[LayerMC]

    def get_dict(self) -> dict[str, int]:
        return self.model_dump(exclude={"mc_per_layer"})

    def get_json(self) -> str:
        return json.dumps(self.get_dict())

    def get_summary(self) -> str:
        summary = (
            f"\nCodebase MC: {self.maintainability_cost:.2f} "
            f"({self.raw_line_cost:.2f} raw MC, without size penalty)\n"
        )
        for layer in self.mc_per_layer:
            summary += (
                f"  Layer {layer.name}: "
                f"{layer.maintainability_cost:.2f} ({layer.raw_line_cost:.2f} raw)\n"
            )
        return summary


class MaintainabilityCostMetric(Metric[MaintainabilityCostResults]): ...



# pylint: disable=invalid-name

import ast

# This has been based on Radon's implementation and updated to support
# modern python features.


class CCVisitor(ast.NodeVisitor):
    def __init__(self) -> None:
        self.complexity = 0  # no CC will be counted per file, only per application

    # Selection statements

    def visit_If(self, node: ast.If) -> None:
        self.complexity += 1
        self.generic_visit(node)

    def visit_Elif(self, node: ast.If) -> None:
        self.complexity += 1
        self.generic_visit(node)

    def visit_IfExp(self, node: ast.IfExp) -> None:
        self.complexity += 1
        self.generic_visit(node)

    def visit_Match(self, node: ast.Match) -> None:
        self.complexity += 1 + len(node.cases)
        self.generic_visit(node)

    def visit_BoolOp(self, node: ast.BoolOp) -> None:
        if isinstance(node.op, (ast.And, ast.Or)):
            self.complexity += len(node.values) - 1
        self.generic_visit(node)

    # Iteration statements

    def visit_For(self, node: ast.For) -> None:
        self.complexity += 1 + bool(node.orelse)
        self.generic_visit(node)

    def visit_AsyncFor(self, node: ast.AsyncFor) -> None:
        self.complexity += 1 + bool(node.orelse)
        self.generic_visit(node)

    def visit_While(self, node: ast.While) -> None:
        self.complexity += 1 + bool(node.orelse)
        self.generic_visit(node)

    # Contexts and exceptions branching

    def visit_Try(self, node: ast.Try) -> None:
        self.complexity += len(node.handlers) + bool(node.orelse)
        self.generic_visit(node)

    def visit_With(self, node: ast.With) -> None:
        self.complexity += len(node.items)
        self.generic_visit(node)

    def visit_AsyncWith(self, node: ast.AsyncWith) -> None:
        self.complexity += len(node.items)
        self.generic_visit(node)

    def visit_Assert(self, node: ast.Assert) -> None:
        self.complexity += 1
        self.generic_visit(node)

    # Comprehensions

    def _visit_comprehensions(self, generators: list[ast.comprehension]) -> None:
        self.complexity += len(generators)
        for gen in generators:
            if gen.ifs:
                self.complexity += len(gen.ifs)

    def visit_ListComp(self, node: ast.ListComp) -> None:
        self._visit_comprehensions(node.generators)
        self.generic_visit(node)

    def visit_SetComp(self, node: ast.SetComp) -> None:
        self._visit_comprehensions(node.generators)
        self.generic_visit(node)

    def visit_DictComp(self, node: ast.DictComp) -> None:
        self._visit_comprehensions(node.generators)
        self.generic_visit(node)

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> None:
        self._visit_comprehensions(node.generators)
        self.generic_visit(node)
